"use strict";


/**
 * Base frame engine
 * @param {string} canvas         The canvas id in which we'll draw
 * @param {function} updateCallback update callback
 * @param {function} renderCallback render callback
 */
function FrameEngine(canvas, updateCallback, renderCallback) {
  this.updateCallback = updateCallback;
  this.renderCallback = renderCallback;
  this.canvas = canvas;
  // Get a 2D Context
  this.context = document.getElementById(canvas).getContext('2d');
  // Detect best timestamp function
  this.timestamp = window.performance && window.performance.now ? this.timestampPerf : this.timestampDate;
  
  // Marker to stop main game loop
  this.exitLoop = true;
}

/**
 * Timestamp fallback function if window.performance is not available
 * @return {integer} a timestamp
 */
FrameEngine.prototype.timestampDate = function() {
  return new Date().getTime();
};

/**
 * Timestamp function
 * @return {integer} a timestamp
 */
FrameEngine.prototype.timestampPerf = function() {
  return window.performance.now();
};


/**
 * Start the game loop
 */
FrameEngine.prototype.start = function () {
  this.last = this.timestamp();
  this.step = 1/30;
  this.exitLoop = false;
  this.dt = 0;

  var self = this;
  requestAnimationFrame(function() { self.gameLoop(); });
};

/**
 * Stop the game loop
 */
FrameEngine.prototype.stop = function() {
  this.exitLoop = true;
};


/**
 * Main Game loop, called constantly by the browser
 */
FrameEngine.prototype.gameLoop = function () {
  var now = this.timestamp();
  // Add elapsed time since last call 
  this.dt = this.dt + Math.min(1, (now - this.last) / 1000);

  // Here we call the update method at given interval
  // catching up time if necessary
  while(this.dt > this.step) {
    this.dt = this.dt - this.step;
    this.update(this.step);
  }

  // Render the scene
  this.render(this.context, this.dt);
  this.last = now;

  // Request the browser to call us again when ready
  if(!this.exitLoop) {
    var self = this;
    requestAnimationFrame(function() { self.gameLoop(); });
  }
};

/**
 * Update game environment
 * @param  {int} dt time elapsed
 */
FrameEngine.prototype.update = function(dt) {
  this.updateCallback(dt);  
};

/**
 * Render game scene  
 * @param  {2DContext} context
 * @param  {int} dt      elapsed time
 */
FrameEngine.prototype.render = function(context, dt) {
  // Clear canvas before rendering
  this.context.clearRect(0,0, this.canvas.width, this.canvas.height);

  this.renderCallback(context, dt);  
};
